#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2016, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.
from __future__ import annotations

from typing import TYPE_CHECKING

from oc_ocdm.abstract_entity import AbstractEntity
from rdflib import RDF, Graph, Namespace, URIRef

if TYPE_CHECKING:
    from typing import ClassVar, Dict, List, Optional, Tuple

    from oc_ocdm.graph.graph_set import GraphSet


class GraphEntity(AbstractEntity):
    BIRO: ClassVar[Namespace] = Namespace("http://purl.org/spar/biro/")
    C4O: ClassVar[Namespace] = Namespace("http://purl.org/spar/c4o/")
    CO: ClassVar[Namespace] = Namespace("http://purl.org/co/")
    CITO: ClassVar[Namespace] = Namespace("http://purl.org/spar/cito/")
    DATACITE: ClassVar[Namespace] = Namespace("http://purl.org/spar/datacite/")
    DCTERMS: ClassVar[Namespace] = Namespace("http://purl.org/dc/terms/")
    DEO: ClassVar[Namespace] = Namespace("http://purl.org/spar/deo/")
    DOCO: ClassVar[Namespace] = Namespace("http://purl.org/spar/doco/")
    FABIO: ClassVar[Namespace] = Namespace("http://purl.org/spar/fabio/")
    FOAF: ClassVar[Namespace] = Namespace("http://xmlns.com/foaf/0.1/")
    FR: ClassVar[Namespace] = Namespace("http://purl.org/spar/fr/")
    FRBR: ClassVar[Namespace] = Namespace("http://purl.org/vocab/frbr/core#")
    LITERAL: ClassVar[Namespace] = Namespace("http://www.essepuntato.it/2010/06/literalreification/")
    OA: ClassVar[Namespace] = Namespace("http://www.w3.org/ns/oa#")
    OCO: ClassVar[Namespace] = Namespace("https://w3id.org/oc/ontology/")
    PRISM: ClassVar[Namespace] = Namespace("http://prismstandard.org/namespaces/basic/2.0/")
    PRO: ClassVar[Namespace] = Namespace("http://purl.org/spar/pro/")

    iri_has_subtitle: ClassVar[URIRef] = FABIO.hasSubtitle
    iri_has_publication_date: ClassVar[URIRef] = PRISM.publicationDate
    iri_bibliographic_reference: ClassVar[URIRef] = BIRO.BibliographicReference
    iri_references: ClassVar[URIRef] = BIRO.references
    iri_denotes: ClassVar[URIRef] = C4O.denotes
    iri_has_content: ClassVar[URIRef] = C4O.hasContent
    iri_intextref_pointer: ClassVar[URIRef] = C4O.InTextReferencePointer
    iri_is_context_of: ClassVar[URIRef] = C4O.isContextOf
    iri_singleloc_pointer_list: ClassVar[URIRef] = C4O.SingleLocationPointerList
    iri_has_element: ClassVar[URIRef] = CO.element
    iri_citation: ClassVar[URIRef] = CITO.Citation
    iri_cites: ClassVar[URIRef] = CITO.cites
    iri_citation_characterisation: ClassVar[URIRef] = CITO.hasCitationCharacterisation
    iri_has_citing_entity: ClassVar[URIRef] = CITO.hasCitingEntity
    iri_has_cited_entity: ClassVar[URIRef] = CITO.hasCitedEntity
    iri_openalex: ClassVar[URIRef] = DATACITE.openalex
    iri_arxiv: ClassVar[URIRef] = DATACITE.arxiv
    iri_oci: ClassVar[URIRef] = DATACITE.oci
    iri_doi: ClassVar[URIRef] = DATACITE.doi
    iri_pmid: ClassVar[URIRef] = DATACITE.pmid
    iri_pmcid: ClassVar[URIRef] = DATACITE.pmcid
    iri_orcid: ClassVar[URIRef] = DATACITE.orcid
    iri_xpath: ClassVar[URIRef] = DATACITE["local-resource-identifier-scheme"]
    iri_intrepid: ClassVar[URIRef] = DATACITE.intrepid
    iri_xmlid: ClassVar[URIRef] = DATACITE["local-resource-identifier-scheme"]
    iri_has_identifier: ClassVar[URIRef] = DATACITE.hasIdentifier
    iri_identifier: ClassVar[URIRef] = DATACITE.Identifier
    iri_isbn: ClassVar[URIRef] = DATACITE.isbn
    iri_issn: ClassVar[URIRef] = DATACITE.issn
    iri_url: ClassVar[URIRef] = DATACITE.url
    iri_uses_identifier_scheme: ClassVar[URIRef] = DATACITE.usesIdentifierScheme
    iri_title: ClassVar[URIRef] = DCTERMS["title"]
    iri_caption: ClassVar[URIRef] = DEO.Caption
    iri_discourse_element: ClassVar[URIRef] = DEO.DiscourseElement
    iri_footnote: ClassVar[URIRef] = DOCO.Footnote
    iri_paragraph: ClassVar[URIRef] = DOCO.Paragraph
    iri_part: ClassVar[URIRef] = DOCO.Part
    iri_section: ClassVar[URIRef] = DOCO.Section
    iri_introduction: ClassVar[URIRef] = DEO.Introduction
    iri_methods: ClassVar[URIRef] = DEO.Methods
    iri_materials: ClassVar[URIRef] = DEO.Materials
    iri_related_work: ClassVar[URIRef] = DEO.RelatedWork
    iri_results: ClassVar[URIRef] = DEO.Results
    iri_discussion: ClassVar[URIRef] = DEO.Discussion
    iri_conclusion: ClassVar[URIRef] = DEO.Conclusion
    iri_section_title: ClassVar[URIRef] = DOCO.SectionTitle
    iri_sentence: ClassVar[URIRef] = DOCO.Sentence
    iri_table: ClassVar[URIRef] = DOCO.Table
    iri_text_chunk: ClassVar[URIRef] = DOCO.TextChunk
    iri_abstract: ClassVar[URIRef] = DOCO.Abstract
    iri_academic_proceedings: ClassVar[URIRef] = FABIO.AcademicProceedings
    iri_audio_document: ClassVar[URIRef] = FABIO.AudioDocument
    iri_book: ClassVar[URIRef] = FABIO.Book
    iri_book_chapter: ClassVar[URIRef] = FABIO.BookChapter
    iri_book_series: ClassVar[URIRef] = FABIO.BookSeries
    iri_book_set: ClassVar[URIRef] = FABIO.BookSet
    iri_computer_program: ClassVar[URIRef] = FABIO.ComputerProgram
    iri_data_file: ClassVar[URIRef] = FABIO.DataFile
    iri_data_management_plan: ClassVar[URIRef] = FABIO.DataManagementPlan
    iri_editorial: ClassVar[URIRef] = FABIO.Editorial
    iri_expression: ClassVar[URIRef] = FABIO.Expression
    iri_expression_collection: ClassVar[URIRef] = FABIO.ExpressionCollection
    iri_has_sequence_identifier: ClassVar[URIRef] = FABIO.hasSequenceIdentifier
    iri_journal: ClassVar[URIRef] = FABIO.Journal
    iri_journal_article: ClassVar[URIRef] = FABIO.JournalArticle
    iri_journal_editorial: ClassVar[URIRef] = FABIO.JournalEditorial
    iri_journal_issue: ClassVar[URIRef] = FABIO.JournalIssue
    iri_journal_volume: ClassVar[URIRef] = FABIO.JournalVolume
    iri_manifestation: ClassVar[URIRef] = FABIO.Manifestation
    iri_newspaper: ClassVar[URIRef] = FABIO.Newspaper
    iri_newspaper_article: ClassVar[URIRef] = FABIO.NewspaperArticle
    iri_newspaper_editorial: ClassVar[URIRef] = FABIO.NewspaperEditorial
    iri_newspaper_issue: ClassVar[URIRef] = FABIO.NewspaperIssue
    iri_peer_review: ClassVar[URIRef] = FR.ReviewVersion
    iri_preprint: ClassVar[URIRef] = FABIO.Preprint
    iri_presentation: ClassVar[URIRef] = FABIO.Presentation
    iri_proceedings_paper: ClassVar[URIRef] = FABIO.ProceedingsPaper
    iri_proceedings_series: ClassVar[URIRef] = FABIO.Series
    iri_reference_book: ClassVar[URIRef] = FABIO.ReferenceBook
    iri_reference_entry: ClassVar[URIRef] = FABIO.ReferenceEntry
    iri_report_document: ClassVar[URIRef] = FABIO.ReportDocument
    iri_retraction_notice: ClassVar[URIRef] = FABIO.RetractionNotice
    iri_series: ClassVar[URIRef] = FABIO.Series
    iri_specification_document: ClassVar[URIRef] = FABIO.SpecificationDocument
    iri_thesis: ClassVar[URIRef] = FABIO.Thesis
    iri_web_content: ClassVar[URIRef] = FABIO.WebContent
    iri_agent: ClassVar[URIRef] = FOAF.Agent
    iri_family_name: ClassVar[URIRef] = FOAF.familyName
    iri_given_name: ClassVar[URIRef] = FOAF.givenName
    iri_name: ClassVar[URIRef] = FOAF.name
    iri_embodiment: ClassVar[URIRef] = FRBR.embodiment
    iri_part_of: ClassVar[URIRef] = FRBR.partOf
    iri_contains_reference: ClassVar[URIRef] = FRBR.part
    iri_contains_de: ClassVar[URIRef] = FRBR.part
    iri_has_literal_value: ClassVar[URIRef] = LITERAL.hasLiteralValue
    iri_ending_page: ClassVar[URIRef] = PRISM.endingPage
    iri_starting_page: ClassVar[URIRef] = PRISM.startingPage
    iri_author: ClassVar[URIRef] = PRO.author
    iri_editor: ClassVar[URIRef] = PRO.editor
    iri_is_held_by: ClassVar[URIRef] = PRO.isHeldBy
    iri_publisher: ClassVar[URIRef] = PRO.publisher
    iri_is_document_context_for: ClassVar[URIRef] = PRO.isDocumentContextFor
    iri_role_in_time: ClassVar[URIRef] = PRO.RoleInTime
    iri_with_role: ClassVar[URIRef] = PRO.withRole
    iri_note: ClassVar[URIRef] = OA.Annotation
    iri_has_body: ClassVar[URIRef] = OA.hasBody
    iri_has_annotation: ClassVar[URIRef] = OCO.hasAnnotation  # inverse of OA.hasTarget
    iri_has_next: ClassVar[URIRef] = OCO.hasNext
    iri_archival_document: ClassVar[URIRef] = FABIO.ArchivalDocument
    iri_viaf: ClassVar[URIRef] = DATACITE.viaf
    iri_crossref: ClassVar[URIRef] = DATACITE.crossref  # TODO: add to datacite!
    iri_datacite: ClassVar[URIRef] = DATACITE.datacite  # TODO: add to datacite!
    iri_jid: ClassVar[URIRef] = DATACITE.jid # TODO: add to datacite!
    iri_wikidata: ClassVar[URIRef] = DATACITE.wikidata  # TODO: add to datacite!
    iri_wikipedia: ClassVar[URIRef] = DATACITE.wikipedia  # TODO: add to datacite!
    iri_has_edition: ClassVar[URIRef] = PRISM.edition
    iri_relation: ClassVar[URIRef] = DCTERMS.relation
    iri_has_citation_creation_date: ClassVar[URIRef] = CITO.hasCitationCreationDate
    iri_has_citation_time_span: ClassVar[URIRef] = CITO.hasCitationTimeSpan
    iri_digital_manifestation: ClassVar[URIRef] = FABIO.DigitalManifestation
    iri_print_object: ClassVar[URIRef] = FABIO.PrintObject
    iri_has_url: ClassVar[URIRef] = FRBR.exemplar
    iri_self_citation: ClassVar[URIRef] = CITO.SelfCitation
    iri_affiliation_self_citation: ClassVar[URIRef] = CITO.AffiliationSelfCitation
    iri_author_network_self_citation: ClassVar[URIRef] = CITO.AuthorNetworkSelfCitation
    iri_author_self_citation: ClassVar[URIRef] = CITO.AuthorSelfCitation
    iri_funder_self_citation: ClassVar[URIRef] = CITO.FunderSelfCitation
    iri_journal_self_citation: ClassVar[URIRef] = CITO.JournalSelfCitation
    iri_journal_cartel_citation: ClassVar[URIRef] = CITO.JournalCartelCitation
    iri_distant_citation: ClassVar[URIRef] = CITO.DistantCitation
    iri_has_format: ClassVar[URIRef] = DCTERMS["format"]

    short_name_to_type_iri: ClassVar[Dict[str, URIRef]] = {
        'an': iri_note,
        'ar': iri_role_in_time,
        'be': iri_bibliographic_reference,
        'br': iri_expression,
        'ci': iri_citation,
        'de': iri_discourse_element,
        'id': iri_identifier,
        'pl': iri_singleloc_pointer_list,
        'ra': iri_agent,
        're': iri_manifestation,
        'rp': iri_intextref_pointer
    }

    def __init__(self, g: Graph, g_set: GraphSet, res: URIRef = None, res_type: URIRef = None,
                 resp_agent: str = None, source: str = None, count: str = None, label: str = None,
                 short_name: str = "", preexisting_graph: Graph = None) -> None:
        super(GraphEntity, self).__init__()
        self.g: Graph = g
        self.resp_agent: str = resp_agent
        self.source: str = source
        self.short_name: str = short_name
        self.g_set: GraphSet = g_set
        self.preexisting_graph: Graph = Graph(identifier=g.identifier)
        self._merge_list: Tuple[GraphEntity] = tuple()
        # FLAGS
        self._to_be_deleted: bool = False
        self._was_merged: bool = False
        self._is_restored: bool = False

        # If res was not specified, create from scratch the URI reference for this entity,
        # otherwise use the provided one
        if res is None:
            self.res = self._generate_new_res(g, count)
        else:
            self.res = res

        if g_set is not None:
            # If not already done, register this GraphEntity instance inside the GraphSet
            if self.res not in g_set.res_to_entity:
                g_set.res_to_entity[self.res] = self

        if preexisting_graph is not None:
            # Triples inside self.g are entirely replaced by triples from preexisting_graph.
            # This has maximum priority with respect to every other self.g initializations.
            # It's fundamental that the preexisting graph gets passed as an argument of the constructor:
            # allowing the user to set this value later through a method would mean that the user could
            # set the preexisting graph AFTER having modified self.g (which would not make sense).
            self.remove_every_triple()
            for p, o in preexisting_graph.predicate_objects(self.res):
                self.g.add((self.res, p, o))
                self.preexisting_graph.add((self.res, p, o))
        else:
            # Add mandatory information to the entity graph
            self._create_type(res_type)
            if label is not None:
                self.create_label(label)

    @staticmethod
    def _generate_new_res(g: Graph, count: str) -> URIRef:
        return URIRef(str(g.identifier) + count)

    @property
    def to_be_deleted(self) -> bool:
        return self._to_be_deleted

    @property
    def was_merged(self) -> bool:
        return self._was_merged

    @property
    def merge_list(self) -> Tuple[GraphEntity]:
        return self._merge_list

    @property 
    def is_restored(self) -> bool:
        """Indicates if this entity was restored after being deleted."""
        return self._is_restored

    def mark_as_restored(self) -> None:
        """
        Marks an entity as being restored after deletion.
                
        This state signals to the provenance system that:
        - No new invalidation time should be generated for the previous snapshot
        - The original deletion snapshot's invalidation time should be preserved
        - The entity should be treated as restored rather than newly created
        """
        self._to_be_deleted = False
        self._is_restored = True
            
    def mark_as_to_be_deleted(self) -> None:
        # Here we must REMOVE triples pointing
        # to 'self' [THIS CANNOT BE UNDONE]:
        for res, entity in self.g_set.res_to_entity.items():
            triples_list: List[Tuple] = list(entity.g.triples((res, None, self.res)))
            for triple in triples_list:
                entity.g.remove(triple)

        self._to_be_deleted = True

    def _get_specific_type(self) -> Optional[URIRef]:
        """
        Get the specific type of the entity (e.g., JournalArticle), if any.
        Excludes the base Expression type.
        
        Returns:
            The specific type URI if present, None otherwise
        """
        base_type = self.short_name_to_type_iri[self.short_name]
        for _, _, type_uri in self.g.triples((self.res, RDF.type, None)):
            if type_uri != base_type:
                return type_uri
        return None

    def merge(self, other: GraphEntity, prefer_self: bool = False) -> None:
        """
        **WARNING:** ``GraphEntity`` **is an abstract class that cannot be instantiated at runtime.
        As such, it's only possible to execute this method on entities generated from**
        ``GraphEntity``'s **subclasses. Please, refer to their documentation of the** `merge` **method.**

        :param other: The entity which will be marked as to be deleted and whose properties will
         be merged into the current entity.
        :type other: GraphEntity
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """

        # Here we must REDIRECT triples pointing
        # to 'other' to make them point to 'self':
        for res, entity in self.g_set.res_to_entity.items():
            triples_list: List[Tuple] = list(entity.g.triples((res, None, other.res)))
            for triple in triples_list:
                entity.g.remove(triple)
                new_triple = (triple[0], triple[1], self.res)
                entity.g.add(new_triple)

        self_specific_type = self._get_specific_type()
        other_specific_type = other._get_specific_type()

        final_specific_type = None
        if prefer_self and self_specific_type:
            final_specific_type = self_specific_type
        elif other_specific_type:
            final_specific_type = other_specific_type
        elif self_specific_type:
            final_specific_type = self_specific_type

        self.g.remove((self.res, RDF.type, None))
        base_type = self.short_name_to_type_iri[self.short_name]
        self.g.add((self.res, RDF.type, base_type))
        if final_specific_type:
            self.g.add((self.res, RDF.type, final_specific_type))

        label: Optional[str] = other.get_label()
        if label is not None:
            self.create_label(label)

        self._was_merged = True
        self._merge_list = (*self._merge_list, other)

        # 'other' must be deleted AFTER the redirection of
        # triples pointing to it, since mark_as_to_be_deleted
        # also removes every triple pointing to 'other'
        other.mark_as_to_be_deleted()

    def commit_changes(self):
        self.preexisting_graph = Graph(identifier=self.g.identifier)
        if self._to_be_deleted:
            self.remove_every_triple()
        else:
            for triple in self.g.triples((self.res, None, None)):
                self.preexisting_graph.add(triple)
        self._is_restored = False
        self._to_be_deleted = False
        self._was_merged = False
        self._merge_list = tuple()


#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2016, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.
from __future__ import annotations

import json
import os
from typing import TYPE_CHECKING
from zipfile import ZipFile

from oc_ocdm.graph.graph_entity import GraphEntity
from oc_ocdm.support.reporter import Reporter
from oc_ocdm.support.support import build_graph_from_results
from rdflib import RDF, Dataset, Graph, URIRef
from sparqlite import SPARQLClient, EndpointError

if TYPE_CHECKING:
    from typing import Any, Dict, List, Optional, Set
    from oc_ocdm.graph.graph_set import GraphSet

from pyshacl import validate


class Reader(object):

    def __init__(self, repok: Reporter = None, reperr: Reporter = None, context_map: Dict[str, Any] = None) -> None:

        if context_map is not None:
            self.context_map: Dict[str, Any] = context_map
        else:
            self.context_map: Dict[str, Any] = {}
        for context_url in self.context_map:
            ctx_file_path: Any = self.context_map[context_url]
            if type(ctx_file_path) == str and os.path.isfile(ctx_file_path):
                # This expensive operation is done only when it's really needed
                with open(ctx_file_path, 'rt', encoding='utf-8') as ctx_f:
                    self.context_map[context_url] = json.load(ctx_f)

        if repok is None:
            self.repok: Reporter = Reporter(prefix="[Reader: INFO] ")
        else:
            self.repok: Reporter = repok

        if reperr is None:
            self.reperr: Reporter = Reporter(prefix="[Reader: ERROR] ")
        else:
            self.reperr: Reporter = reperr

    def load(self, rdf_file_path: str) -> Optional[Dataset]:
        self.repok.new_article()
        self.reperr.new_article()

        loaded_graph: Optional[Dataset] = None
        if os.path.isfile(rdf_file_path):

            try:
                loaded_graph = self._load_graph(rdf_file_path)
            except Exception as e:
                self.reperr.add_sentence("[1] "
                                         "It was impossible to handle the format used for "
                                         "storing the file (stored in the temporary path) "
                                         f"'{rdf_file_path}'. Additional details: {e}")
        else:
            self.reperr.add_sentence("[2] "
                                     f"The file specified ('{rdf_file_path}') doesn't exist.")

        return loaded_graph

    def _load_graph(self, file_path: str) -> Dataset:
        formats = ["json-ld", "rdfxml", "turtle", "trig", "nt11", "nquads"]
        loaded_graph = Dataset()

        if file_path.endswith('.zip'):
            try:
                with ZipFile(file=file_path, mode="r") as archive:
                    for zf_name in archive.namelist():
                        with archive.open(zf_name) as f:
                            if self._try_parse(loaded_graph, f, formats):
                                return loaded_graph
            except Exception as e:
                raise IOError(f"Error opening or reading zip file '{file_path}': {e}")
        else:
            try:
                with open(file_path, 'rt', encoding='utf-8') as f:
                    if self._try_parse(loaded_graph, f, formats):
                        return loaded_graph
            except Exception as e:
                raise IOError(f"Error opening or reading file '{file_path}': {e}")

        raise IOError(f"It was impossible to load the file '{file_path}' with supported formats.")

    def _try_parse(self, graph: Dataset, file_obj, formats: List[str]) -> bool:
        for cur_format in formats:
            file_obj.seek(0)  # Reset file pointer to the beginning for each new attempt
            try:
                if cur_format == "json-ld":
                    json_ld_file = json.load(file_obj)
                    if isinstance(json_ld_file, dict):
                        json_ld_file = [json_ld_file]
                    for json_ld_resource in json_ld_file:
                        if "@context" in json_ld_resource and json_ld_resource["@context"] in self.context_map:
                            json_ld_resource["@context"] = self.context_map[json_ld_resource["@context"]]["@context"]
                    data = json.dumps(json_ld_file, ensure_ascii=False)
                    graph.parse(data=data, format=cur_format)
                else:
                    graph.parse(file=file_obj, format=cur_format)
                return True  # Success, no need to try other formats
            except Exception as e:
                continue  # Try the next format
        return False  # None of the formats succeeded

    @staticmethod
    def get_graph_from_subject(graph: Graph, subject: URIRef) -> Graph:
        g: Graph = Graph(identifier=graph.identifier)
        for p, o in graph.predicate_objects(subject, unique=True):
            g.add((subject, p, o))
        return g

    @staticmethod
    def _extract_subjects(graph: Graph) -> Set[URIRef]:
        subjects: Set[URIRef] = set()
        for s in graph.subjects(unique=True):
            subjects.add(s)
        return subjects

    def graph_validation(self, graph: Graph, closed: bool = False) -> Graph:
        valid_graph: Graph = Graph(identifier=graph.identifier)
        sg = Graph()
        if closed:
            sg.parse(os.path.join('oc_ocdm', 'resources', 'shacle_closed.ttl'))
        else:
            sg.parse(os.path.join('oc_ocdm', 'resources', 'shacle.ttl'))
        _, report_g, _ = validate(graph,
            shacl_graph=sg,
            ont_graph=None,
            inference=None,
            abort_on_first=False,
            allow_infos=False,
            allow_warnings=False,
            meta_shacl=False,
            advanced=False,
            js=False,
            debug=False)
        invalid_nodes = set()
        for triple in report_g.triples((None, URIRef('http://www.w3.org/ns/shacl#focusNode'), None)):
            invalid_nodes.add(triple[2])
        for subject in self._extract_subjects(graph):
            if subject not in invalid_nodes:
                for valid_subject_triple in graph.triples((subject, None, None)):
                    valid_graph.add(valid_subject_triple)
        return valid_graph

    @staticmethod
    def import_entities_from_graph(g_set: GraphSet, results: List[Dict]|Graph|Dataset, resp_agent: str,
                                   enable_validation: bool = False, closed: bool = False) -> List[GraphEntity]:
        if isinstance(results, list):
            graph = build_graph_from_results(results)
        elif isinstance(results, Dataset):
            # Convert Dataset to Graph by flattening all quads into triples
            graph = Graph()
            for s, p, o, _ in results.quads((None, None, None, None)):
                graph.add((s, p, o))
        else:
            graph = results
        if enable_validation:
            reader = Reader()
            graph = reader.graph_validation(graph, closed)
        imported_entities: List[GraphEntity] = []
        for subject in Reader._extract_subjects(graph):
            types = []
            for o in graph.objects(subject, RDF.type):
                types.append(o)
            # ReferenceAnnotation
            if GraphEntity.iri_note in types:
                imported_entities.append(g_set.add_an(resp_agent=resp_agent, res=subject,
                                         preexisting_graph=Reader.get_graph_from_subject(graph, subject)))
            # AgentRole
            elif GraphEntity.iri_role_in_time in types:
                imported_entities.append(g_set.add_ar(resp_agent=resp_agent, res=subject,
                                         preexisting_graph=Reader.get_graph_from_subject(graph, subject)))
            # BibliographicReference
            elif GraphEntity.iri_bibliographic_reference in types:
                imported_entities.append(g_set.add_be(resp_agent=resp_agent, res=subject,
                                         preexisting_graph=Reader.get_graph_from_subject(graph, subject)))
            # BibliographicResource
            elif GraphEntity.iri_expression in types:
                imported_entities.append(g_set.add_br(resp_agent=resp_agent, res=subject,
                                         preexisting_graph=Reader.get_graph_from_subject(graph, subject)))
            # Citation
            elif GraphEntity.iri_citation in types:
                imported_entities.append(g_set.add_ci(resp_agent=resp_agent, res=subject,
                                         preexisting_graph=Reader.get_graph_from_subject(graph, subject)))
            # DiscourseElement
            elif GraphEntity.iri_discourse_element in types:
                imported_entities.append(g_set.add_de(resp_agent=resp_agent, res=subject,
                                         preexisting_graph=Reader.get_graph_from_subject(graph, subject)))
            # Identifier
            elif GraphEntity.iri_identifier in types:
                imported_entities.append(g_set.add_id(resp_agent=resp_agent, res=subject,
                                         preexisting_graph=Reader.get_graph_from_subject(graph, subject)))
            # PointerList
            elif GraphEntity.iri_singleloc_pointer_list in types:
                imported_entities.append(g_set.add_pl(resp_agent=resp_agent, res=subject,
                                         preexisting_graph=Reader.get_graph_from_subject(graph, subject)))
            # ResponsibleAgent
            elif GraphEntity.iri_agent in types:
                imported_entities.append(g_set.add_ra(resp_agent=resp_agent, res=subject,
                                         preexisting_graph=Reader.get_graph_from_subject(graph, subject)))
            # ResourceEmbodiment
            elif GraphEntity.iri_manifestation in types:
                imported_entities.append(g_set.add_re(resp_agent=resp_agent, res=subject,
                                         preexisting_graph=Reader.get_graph_from_subject(graph, subject)))
            # ReferencePointer
            elif GraphEntity.iri_intextref_pointer in types:
                imported_entities.append(g_set.add_rp(resp_agent=resp_agent, res=subject,
                                         preexisting_graph=Reader.get_graph_from_subject(graph, subject)))
        return imported_entities

    @staticmethod
    def import_entity_from_triplestore(g_set: GraphSet, ts_url: str, res: URIRef, resp_agent: str,
                                    enable_validation: bool = False) -> GraphEntity:
        query: str = f"SELECT ?s ?p ?o WHERE {{BIND (<{res}> AS ?s). ?s ?p ?o.}}"
        try:
            with SPARQLClient(ts_url, max_retries=3, backoff_factor=2.5) as client:
                result = client.query(query)['results']['bindings']

                if not result:
                    raise ValueError(f"The requested entity {res} was not found in the triplestore.")

                imported_entities: List[GraphEntity] = Reader.import_entities_from_graph(g_set, result, resp_agent, enable_validation)
                if len(imported_entities) <= 0:
                    raise ValueError("The requested entity was not recognized as a proper OCDM entity.")
                return imported_entities[0]

        except ValueError:
            raise
        except EndpointError as e:
            print(f"[3] Could not import entity due to communication problems: {e}")
            raise

    @staticmethod
    def import_entities_from_triplestore(g_set: GraphSet, ts_url: str, entities: List[URIRef], resp_agent: str,
                                    enable_validation: bool = False, batch_size: int = 1000) -> List[GraphEntity]:
        if not entities:
            raise ValueError("No entities provided for import")

        imported_entities: List[GraphEntity] = []

        try:
            with SPARQLClient(ts_url, max_retries=3, backoff_factor=2.5) as client:
                for i in range(0, len(entities), batch_size):
                    batch = entities[i:i + batch_size]
                    not_found_entities = set(str(entity) for entity in batch)

                    union_patterns = []
                    for entity in batch:
                        union_patterns.append(f"{{ BIND(<{str(entity)}> AS ?s) ?s ?p ?o }}")

                    query = f"""
                    SELECT ?s ?p ?o
                    WHERE {{
                        {' UNION '.join(union_patterns)}
                    }}
                    """

                    results = client.query(query)['results']['bindings']

                    if not results:
                        entities_str = ', '.join(not_found_entities)
                        raise ValueError(f"The requested entities were not found in the triplestore: {entities_str}")

                    for result in results:
                        if 's' in result and 'value' in result['s']:
                            not_found_entities.discard(result['s']['value'])

                    batch_entities = Reader.import_entities_from_graph(
                        g_set=g_set,
                        results=results,
                        resp_agent=resp_agent,
                        enable_validation=enable_validation
                    )
                    imported_entities.extend(batch_entities)

                    if not_found_entities:
                        entities_str = ', '.join(not_found_entities)
                        raise ValueError(f"The following entities were not recognized as proper OCDM entities: {entities_str}")

        except ValueError:
            raise
        except EndpointError as e:
            print(f"[3] Could not import batch due to communication problems: {e}")
            raise

        if not imported_entities:
            raise ValueError("None of the requested entities were found or recognized as proper OCDM entities.")

        return imported_entities


#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2016, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from oc_ocdm.abstract_entity import AbstractEntity

if TYPE_CHECKING:
    from typing import List, ClassVar, Dict, Optional
    from rdflib import URIRef, Graph


class AbstractSet(ABC):
    """
    Abstract class which represents a generic set of entities.
    It is the base class for each concrete set of entities.
    """

    short_name_to_type_iri: ClassVar[Dict[str, URIRef]] = {}

    def __init__(self) -> None:
        """
        Constructor of the ``AbstractSet`` class.
        """
        self.res_to_entity: Dict[URIRef, AbstractEntity] = {}

    def graphs(self) -> List[Graph]:
        """
        A utility method that allows to retrieve the list of ``rdflib.Graph``
        instances corresponding to each entity contained in the set.

        :return: The requested list of graphs
        """
        result: List[Graph] = []
        for entity in self.res_to_entity.values():
            if len(entity.g) > 0:
                result.append(entity.g)
        return result

    def __getstate__(self):
        """
        Support for pickle serialization.

        RDFLib Graph objects contain threading locks that are not directly picklable,
        but Python's pickle module can handle them. This method provides standard
        pickle protocol support for AbstractSet instances.
        """
        return self.__dict__.copy()

    def __setstate__(self, state):
        """
        Support for pickle deserialization.

        Restores the AbstractSet state from a pickled representation.
        """
        self.__dict__.update(state)

    @abstractmethod
    def get_entity(self, res: URIRef) -> Optional[AbstractEntity]:
        """
        Method signature for concrete implementations that allow
        to retrieve a contained entity identified by its URI.

        :param res: The URI that identifies the requested entity
        :type res: URIRef
        :return: The requested entity if found, None otherwise
        """
        raise NotImplementedError

    @staticmethod
    def get_graph_iri(g: Graph) -> str:
        """
        A utility method that allows to retrieve the IRI which represents
        the name of a given named graph.

        **NOTE: this is a static function!**

        :param g: The named graph whose name will be returned
        :type g: Graph
        :return: The requested string whose content is the IRI associated to the given named graph
        """
        return str(g.identifier)



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2016, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.
from __future__ import annotations

from typing import TYPE_CHECKING

from oc_ocdm.decorators import accepts_only

if TYPE_CHECKING:
    from typing import Optional, List
    from rdflib import URIRef
    from oc_ocdm.graph.entities.bibliographic.reference_annotation import ReferenceAnnotation
    from oc_ocdm.graph.entities.bibliographic.bibliographic_resource import BibliographicResource
from oc_ocdm.graph.graph_entity import GraphEntity
from oc_ocdm.graph.entities.bibliographic_entity import BibliographicEntity


class BibliographicReference(BibliographicEntity):
    """Bibliographic reference (short: be): the particular textual bibliographic reference,
    usually occurring in the reference list (and denoted by one or more in-text reference
    pointers within the text of a citing bibliographic resource), that references another
    bibliographic resource.
    """

    @accepts_only('be')
    def merge(self, other: BibliographicReference) -> None:
        """
        The merge operation allows combining two ``BibliographicReference`` entities into a single one,
        by marking the second entity as to be deleted while also copying its data into the current
        ``BibliographicReference``. Moreover, every triple from the containing ``GraphSet`` referring to the second
        entity gets "redirected" to the current entity: **every other reference contained inside a
        different source (e.g. a triplestore) must be manually handled by the user!**

        In case of functional properties, values from the current entity get overwritten
        by those coming from the second entity while, in all other cases, values from the
        second entity are simply appended to those of the current entity. In this context,
        ``rdfs:label`` is considered as a functional property, while ``rdf:type`` is not.

        :param other: The entity which will be marked as to be deleted and whose properties will
         be merged into the current entity.
        :type other: BibliographicReference
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        super(BibliographicReference, self).merge(other)

        content: Optional[str] = other.get_content()
        if content is not None:
            self.has_content(content)

        annotations_list: List[ReferenceAnnotation] = other.get_annotations()
        for cur_annotation in annotations_list:
            self.has_annotation(cur_annotation)

        referenced_br: Optional[BibliographicResource] = other.get_referenced_br()
        if referenced_br is not None:
            self.references_br(referenced_br)

    # HAS BIBLIOGRAPHIC REFERENCE TEXT
    def get_content(self) -> Optional[str]:
        """
        Getter method corresponding to the ``c4o:hasContent`` RDF predicate.

        :return: The requested value if found, None otherwise
        """
        return self._get_literal(GraphEntity.iri_has_content)

    @accepts_only('literal')
    def has_content(self, string: str) -> None:
        """
        Setter method corresponding to the ``c4o:hasContent`` RDF predicate.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        `The literal text of a bibliographic reference occurring in the reference list (or
        elsewhere) within a bibliographic resource, that references another bibliographic
        resource. The reference text should be recorded “as given” in the citing bibliographic
        resource, including any errors (e.g. mis-spellings of authors’ names, or changes from
        “β” in the original published title to “beta” in the reference text) or omissions (e.g.
        omission of the title of the referenced bibliographic resource, or omission of sixth and
        subsequent authors’ names, as required by certain publishers), and in whatever format
        it has been made available. For instance, the reference text can be either as plain text
        or as a block of XML.`

        :param string: The value that will be set as the object of the property related to this method
        :type string: str
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self.remove_content()
        self._create_literal(GraphEntity.iri_has_content, string)

    def remove_content(self) -> None:
        """
        Remover method corresponding to the ``c4o:hasContent`` RDF predicate.

        :return: None
        """
        self.g.remove((self.res, GraphEntity.iri_has_content, None))

    # HAS ANNOTATION (ReferenceAnnotation)
    def get_annotations(self) -> List[ReferenceAnnotation]:
        """
        Getter method corresponding to the ``oco:hasAnnotation`` RDF predicate.

        :return: A list containing the requested values if found, None otherwise
        """
        uri_list: List[URIRef] = self._get_multiple_uri_references(GraphEntity.iri_has_annotation, 'an')
        result: List[ReferenceAnnotation] = []
        for uri in uri_list:
            result.append(self.g_set.add_an(self.resp_agent, self.source, uri))
        return result

    @accepts_only('an')
    def has_annotation(self, an_res: ReferenceAnnotation) -> None:
        """
        Setter method corresponding to the ``oco:hasAnnotation`` RDF predicate.

        `An annotation characterizing the related citation, in terms of its citation function (the
        reason for that citation).`

        :param an_res: The value that will be set as the object of the property related to this method
        :type an_res: ReferenceAnnotation
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self.g.add((self.res, GraphEntity.iri_has_annotation, an_res.res))

    @accepts_only('an')
    def remove_annotation(self, an_res: ReferenceAnnotation = None) -> None:
        """
        Remover method corresponding to the ``oco:hasAnnotation`` RDF predicate.

        **WARNING: this is a non-functional property, hence, if the parameter
        is None, any existing value will be removed!**

        :param an_res: If not None, the specific object value that will be removed from the property
         related to this method (defaults to None)
        :type an_res: ReferenceAnnotation
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        if an_res is not None:
            self.g.remove((self.res, GraphEntity.iri_has_annotation, an_res.res))
        else:
            self.g.remove((self.res, GraphEntity.iri_has_annotation, None))

    # REFERENCES (BibliographicResource)
    def get_referenced_br(self) -> Optional[BibliographicResource]:
        """
        Getter method corresponding to the ``biro:references`` RDF predicate.

        :return: The requested value if found, None otherwise
        """
        uri: Optional[URIRef] = self._get_uri_reference(GraphEntity.iri_references, 'br')
        if uri is not None:
            return self.g_set.add_br(self.resp_agent, self.source, uri)

    @accepts_only('br')
    def references_br(self, br_res: BibliographicResource) -> None:
        """
        Setter method corresponding to the ``biro:references`` RDF predicate.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        `The bibliographic reference that cites this bibliographic resource.`

        :param br_res: The value that will be set as the object of the property related to this method
        :type br_res: BibliographicResource
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self.remove_referenced_br()
        self.g.add((self.res, GraphEntity.iri_references, br_res.res))

    def remove_referenced_br(self) -> None:
        """
        Remover method corresponding to the ``biro:references`` RDF predicate.

        :return: None
        """
        self.g.remove((self.res, GraphEntity.iri_references, None))



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2016, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.
from __future__ import annotations

from typing import TYPE_CHECKING

from oc_ocdm.decorators import accepts_only

if TYPE_CHECKING:
    from typing import Optional, List
    from rdflib import URIRef
from oc_ocdm.graph.graph_entity import GraphEntity
from oc_ocdm.graph.entities.bibliographic_entity import BibliographicEntity


class ResponsibleAgent(BibliographicEntity):
    """Responsible agent (short: ra): the agent (usually a person or an organisation) having
       a certain role with respect to a bibliographic resource (e.g. an author of a paper or
       book, or an editor of a journal)."""

    @accepts_only('ra')
    def merge(self, other: ResponsibleAgent) -> None:
        """
        The merge operation allows combining two ``ResponsibleAgent`` entities into a single one,
        by marking the second entity as to be deleted while also copying its data into the current
        ``ResponsibleAgent``. Moreover, every triple from the containing ``GraphSet`` referring to the second
        entity gets "redirected" to the current entity: **every other reference contained inside a
        different source (e.g. a triplestore) must be manually handled by the user!**

        In case of functional properties, values from the current entity get overwritten
        by those coming from the second entity while, in all other cases, values from the
        second entity are simply appended to those of the current entity. In this context,
        ``rdfs:label`` is considered as a functional property, while ``rdf:type`` is not.

        :param other: The entity which will be marked as to be deleted and whose properties will
         be merged into the current entity.
        :type other: ResponsibleAgent
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        super(ResponsibleAgent, self).merge(other)

        name: Optional[str] = other.get_name()
        if name is not None:
            self.has_name(name)

        given_name: Optional[str] = other.get_given_name()
        if given_name is not None:
            self.has_given_name(given_name)

        family_name: Optional[str] = other.get_family_name()
        if family_name is not None:
            self.has_family_name(family_name)

        related_agents: List[URIRef] = other.get_related_agents()
        for cur_agent in related_agents:
            self.has_related_agent(cur_agent)

    # HAS NAME
    def get_name(self) -> Optional[str]:
        """
        Getter method corresponding to the ``foaf:name`` RDF predicate.

        :return: The requested value if found, None otherwise
        """
        return self._get_literal(GraphEntity.iri_name)

    @accepts_only('literal')
    def has_name(self, string: str) -> None:
        """
        Setter method corresponding to the ``foaf:name`` RDF predicate.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        `The name of an agent (for people, usually in the format: given name followed by family
        name, separated by a space).`

        :param string: The value that will be set as the object of the property related to this method
        :type string: str
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self.remove_name()
        self._create_literal(GraphEntity.iri_name, string)

    def remove_name(self) -> None:
        """
        Remover method corresponding to the ``foaf:name`` RDF predicate.

        :return: None
        """
        self.g.remove((self.res, GraphEntity.iri_name, None))

    # HAS GIVEN NAME
    def get_given_name(self) -> Optional[str]:
        """
        Getter method corresponding to the ``foaf:givenName`` RDF predicate.

        :return: The requested value if found, None otherwise
        """
        return self._get_literal(GraphEntity.iri_given_name)

    @accepts_only('literal')
    def has_given_name(self, string: str) -> None:
        """
        Setter method corresponding to the ``foaf:givenName`` RDF predicate.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        `The given name of an agent, if a person.`

        :param string: The value that will be set as the object of the property related to this method
        :type string: str
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self.remove_given_name()
        self._create_literal(GraphEntity.iri_given_name, string)

    def remove_given_name(self) -> None:
        """
        Remover method corresponding to the ``foaf:givenName`` RDF predicate.

        :return: None
        """
        self.g.remove((self.res, GraphEntity.iri_given_name, None))

    # HAS FAMILY NAME
    def get_family_name(self) -> Optional[str]:
        """
        Getter method corresponding to the ``foaf:familyName`` RDF predicate.

        :return: The requested value if found, None otherwise
        """
        return self._get_literal(GraphEntity.iri_family_name)

    @accepts_only('literal')
    def has_family_name(self, string: str) -> None:
        """
        Setter method corresponding to the ``foaf:familyName`` RDF predicate.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        `The family name of an agent, if a person.`

        :param string: The value that will be set as the object of the property related to this method
        :type string: str
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self.remove_family_name()
        self._create_literal(GraphEntity.iri_family_name, string)

    def remove_family_name(self) -> None:
        """
        Remover method corresponding to the ``foaf:familyName`` RDF predicate.

        :return: None
        """
        self.g.remove((self.res, GraphEntity.iri_family_name, None))

    # HAS RELATED AGENT
    def get_related_agents(self) -> List[URIRef]:
        """
        Getter method corresponding to the ``dcterms:relation`` RDF predicate.

        :return: A list containing the requested values if found, None otherwise
        """
        uri_list: List[URIRef] = self._get_multiple_uri_references(GraphEntity.iri_relation)
        return uri_list

    @accepts_only('thing')
    def has_related_agent(self, thing_res: URIRef) -> None:
        """
        Setter method corresponding to the ``dcterms:relation`` RDF predicate.

        `An external agent that/who is related in some relevant way with this responsible agent
        (e.g. for inter-linking purposes).`

        :param thing_res: The value that will be set as the object of the property related to this method
        :type thing_res: str
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self.g.add((self.res, GraphEntity.iri_relation, thing_res))

    @accepts_only('thing')
    def remove_related_agent(self, thing_res: URIRef = None) -> None:
        """
        Remover method corresponding to the ``dcterms:relation`` RDF predicate.

        **WARNING: this is a non-functional property, hence, if the parameter
        is None, any existing value will be removed!**

        :param thing_res: If not None, the specific object value that will be removed from the property
         related to this method (defaults to None)
        :type thing_res: URIRef
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        if thing_res is not None:
            self.g.remove((self.res, GraphEntity.iri_relation, thing_res))
        else:
            self.g.remove((self.res, GraphEntity.iri_relation, None))



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2016, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.

from oc_ocdm.support.support import create_date, get_datatype_from_iso_8601, encode_url, create_literal,\
                                    create_type, is_string_empty, get_short_name, get_prefix, get_count,\
                                    get_resource_number, find_local_line_id, find_paths, has_supplier_prefix,\
                                    is_dataset
from oc_ocdm.support.reporter import Reporter
from oc_ocdm.support.query_utils import get_update_query, get_insert_query, get_delete_query



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2016, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.
from __future__ import annotations

from typing import TYPE_CHECKING

from oc_ocdm.counter_handler.counter_handler import CounterHandler
from oc_ocdm.counter_handler.filesystem_counter_handler import FilesystemCounterHandler
from oc_ocdm.counter_handler.in_memory_counter_handler import InMemoryCounterHandler
from oc_ocdm.metadata.entities.dataset import Dataset
from oc_ocdm.metadata.entities.distribution import Distribution
from oc_ocdm.support.support import get_count, is_dataset, get_short_name

if TYPE_CHECKING:
    from typing import Dict, Optional, Tuple, ClassVar

from rdflib import Graph, URIRef

from oc_ocdm.metadata.metadata_entity import MetadataEntity
from oc_ocdm.abstract_set import AbstractSet


class MetadataSet(AbstractSet):
    # Labels
    labels: ClassVar[Dict[str, str]] = {
        "_dataset_": "dataset",
        "di": "distribution"
    }

    def __init__(self, base_iri: str, info_dir: str = "", wanted_label: bool = True) -> None:
        super(MetadataSet, self).__init__()
        # The following variable maps a URIRef with the related metadata entity
        self.res_to_entity: Dict[URIRef, MetadataEntity] = {}
        self.base_iri: str = base_iri
        if self.base_iri[-1] != '/':
            self.base_iri += '/'
        self.wanted_label: bool = wanted_label

        if info_dir is not None and info_dir != "":
            self.counter_handler: CounterHandler = FilesystemCounterHandler(info_dir)
        else:
            self.counter_handler: CounterHandler = InMemoryCounterHandler()

    def get_entity(self, res: URIRef) -> Optional[MetadataEntity]:
        if res in self.res_to_entity:
            return self.res_to_entity[res]

    def add_dataset(self, dataset_name: str, resp_agent: str, source: str = None, res: URIRef = None,
                    preexisting_graph: Graph = None) -> Dataset:
        if res is not None and not is_dataset(res):
            raise ValueError(f"Given res: <{res}> is inappropriate for a Dataset entity.")
        if res is not None and res in self.res_to_entity:
            return self.res_to_entity[res]
        # Here we use a fictitious short name for Dataset, since the OCDM document doesn't specify
        # any particular short name for this type of entity. It's only used internally to distinguish
        # between different metadata entities but it's meaningless outside of this scope.
        cur_g, count, label = self._add_metadata("_dataset_", dataset_name, res)
        return Dataset(cur_g, self.base_iri, dataset_name, self, res,
                       MetadataEntity.iri_dataset, resp_agent,
                       source, count, label, "_dataset_", preexisting_graph)

    def add_di(self, dataset_name: str, resp_agent: str, source: str = None,
               res: URIRef = None, preexisting_graph: Graph = None) -> Distribution:
        if res is not None and get_short_name(res) != "di":
            raise ValueError(f"Given res: <{res}> is inappropriate for a Distribution entity.")
        if res is not None and res in self.res_to_entity:
            return self.res_to_entity[res]
        cur_g, count, label = self._add_metadata("di", dataset_name, res)
        return Distribution(cur_g, self.base_iri, dataset_name, self, res,
                            MetadataEntity.iri_datafile, resp_agent,
                            source, count, label, "di", preexisting_graph)

    def _add_metadata(self, short_name: str, dataset_name: str,
                      res: URIRef = None) -> Tuple[Graph, Optional[str], Optional[str]]:
        cur_g: Graph = Graph()
        self._set_ns(cur_g)

        count: Optional[str] = None
        label: Optional[str] = None

        if res is not None:
            if short_name != '_dataset_':  # Datasets don't have a counter associated with them...
                try:
                    res_count: int = int(get_count(res))
                except ValueError:
                    res_count: int = -1
                if res_count > self.counter_handler.read_metadata_counter(short_name, dataset_name):
                    self.counter_handler.set_metadata_counter(res_count, short_name, dataset_name)
            return cur_g, count, label

        if short_name != '_dataset_':  # Datasets don't have a counter associated with them...
            count = str(self.counter_handler.increment_metadata_counter(short_name, dataset_name))

        if self.wanted_label:
            label = "%s %s [%s/%s]" % (self.labels[short_name], count, short_name, count)

        return cur_g, count, label

    def commit_changes(self):
        for res, entity in self.res_to_entity.items():
            entity.commit_changes()
            if entity.to_be_deleted:
                del self.res_to_entity[res]

    @staticmethod
    def _set_ns(g: Graph) -> None:
        g.namespace_manager.bind("dcterms", MetadataEntity.DCTERMS)
        g.namespace_manager.bind("dcat", MetadataEntity.DCAT)
        g.namespace_manager.bind("void", MetadataEntity.VOID)

    def get_dataset(self) -> Tuple[Dataset]:
        return tuple(entity for entity in self.res_to_entity.values() if isinstance(entity, Dataset))

    def get_di(self) -> Tuple[Distribution]:
        return tuple(entity for entity in self.res_to_entity.values() if isinstance(entity, Distribution))



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2016, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.
from __future__ import annotations

from abc import ABC
from typing import TYPE_CHECKING

from oc_ocdm.support.support import create_type, create_literal, get_short_name, is_dataset, is_string_empty
from rdflib import URIRef, RDFS, RDF, Literal, Graph

if TYPE_CHECKING:
    from typing import Optional, List, ClassVar, Dict, Tuple, Iterable
    from rdflib import term


class AbstractEntity(ABC):
    """
    Abstract class which represents a generic entity from the OCDM. It sits
    at the top of the entity class hierarchy.
    """

    short_name_to_type_iri: ClassVar[Dict[str, URIRef]] = {}

    def __init__(self) -> None:
        """
        Constructor of the ``AbstractEntity`` class.
        """
        self.g: Graph = Graph()
        self.res: URIRef = URIRef("")
        self.short_name: str = ""

    def remove_every_triple(self) -> None:
        """
        Remover method that removes every triple from the current entity.

        **WARNING: the OCDM specification requires that every entity has at least
        one triple that defines its type (through the** ``rdf:type`` **RDF predicate). If
        such triple is not subsequently restored by the user, the entity will be considered
        as to be deleted since it wouldn't be valid anymore.**

        :return: None
        """
        self.g.remove((None, None, None))

    # LABEL
    def get_label(self) -> Optional[str]:
        """
        Getter method corresponding to the ``rdfs:label`` RDF predicate.

        :return: The requested value if found, None otherwise
        """
        return self._get_literal(RDFS.label)

    def create_label(self, string: str) -> None:
        """
        Setter method corresponding to the ``rdfs:label`` RDF predicate.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        :param string: The value that will be set as the object of the property related to this method
        :type string: str
        :return: None
        """
        self.remove_label()
        self._create_literal(RDFS.label, string)

    def remove_label(self) -> None:
        """
        Remover method corresponding to the ``rdfs:label`` RDF predicate.

        :return: None
        """
        self.g.remove((self.res, RDFS.label, None))

    def _create_literal(self, p: URIRef, s: str, dt: URIRef = None, nor: bool = True) -> None:
        """
        Adds an RDF triple with a literal object inside the graph of the entity

        :param p: The predicate
        :type p: URIRef
        :param s: The string to add as a literal value
        :type s: str
        :param dt: The object's datatype, if present
        :type dt: URIRef, optional
        :param nor: Whether to normalize the graph or not
        :type nor: bool, optional
        :return: None
        """
        create_literal(self.g, self.res, p, s, dt, nor)

    # TYPE
    def get_types(self) -> List[URIRef]:
        """
        Getter method corresponding to the ``rdf:type`` RDF predicate.

        :return: A list containing the requested values if found, None otherwise
        """
        uri_list: List[URIRef] = self._get_multiple_uri_references(RDF.type)
        return uri_list

    def _create_type(self, res_type: URIRef) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        :param res_type: The value that will be set as the object of the property related to this method
        :type res_type: URIRef
        :return: None
        """
        self.remove_type()  # <-- It doesn't remove the main type!
        create_type(self.g, self.res, res_type)

    def remove_type(self) -> None:
        """
        Remover method corresponding to the ``rdf:type`` RDF predicate.

        **WARNING: the OCDM specification requires at least one type for an entity.
        This method removes any existing secondary type, without removing the main type.**

        :return: None
        """
        self.g.remove((self.res, RDF.type, None))
        # Restore the main type IRI
        iri_main_type: URIRef = self.short_name_to_type_iri[self.short_name]
        create_type(self.g, self.res, iri_main_type)

    # Overrides __str__ method
    def __str__(self) -> str:
        return str(self.res)

    def add_triples(self, iterable_of_triples: Iterable[Tuple[term]]) -> None:
        """
        A utility method that allows to add a batch of triples into the graph of the entity.

        **WARNING: Only triples that have this entity as their subject will be imported!**

        :param iterable_of_triples: A collection of triples to be added to the entity
        :type iterable_of_triples: Iterable[Tuple[term]]
        :return: None
        """
        for s, p, o in iterable_of_triples:
            if s == self.res:  # This guarantees that only triples belonging to the resource will be added
                self.g.add((s, p, o))

    def _get_literal(self, predicate: URIRef) -> Optional[str]:
        result: Optional[str] = None
        for o in self.g.objects(self.res, predicate):
            if type(o) == Literal:
                result = str(o)
                break
        return result

    def _get_multiple_literals(self, predicate: URIRef) -> List[str]:
        result: List[str] = []
        for o in self.g.objects(self.res, predicate):
            if type(o) == Literal:
                result.append(str(o))
        return result

    def _get_uri_reference(self, predicate: URIRef, short_name: str = None) -> Optional[URIRef]:
        result: Optional[URIRef] = None
        for o in self.g.objects(self.res, predicate):
            if type(o) == URIRef:
                if not is_string_empty(short_name):
                    # If a particular short_name is explicitly requested,
                    # then the following additional check must be performed:
                    if (short_name == '_dataset_' and is_dataset(o)) or get_short_name(o) == short_name:
                        result = o
                        break
                else:
                    result = o
                    break
        return result

    def _get_multiple_uri_references(self, predicate: URIRef, short_name: str = None) -> List[URIRef]:
        result: List[URIRef] = []
        for o in self.g.objects(self.res, predicate):
            if type(o) == URIRef:
                if not is_string_empty(short_name):
                    # If a particular short_name is explicitly requested,
                    # then the following additional check must be performed:
                    if (short_name == '_dataset_' and is_dataset(o)) or get_short_name(o) == short_name:
                        result.append(o)
                else:
                    result.append(o)
        return result



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2016, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.
from __future__ import annotations

from typing import TYPE_CHECKING, Set

if TYPE_CHECKING:
    from typing import Tuple
    from rdflib import URIRef
    from oc_ocdm.abstract_entity import AbstractEntity


def _serialize_triples_to_nt(triples: Set) -> str:
    return "".join(f"{s.n3()} {p.n3()} {o.n3()} ." for s, p, o in triples)


def get_delete_query(graph_iri: URIRef, data: Set) -> Tuple[str, int]:
    num_of_statements: int = len(data)
    if num_of_statements <= 0:
        return "", 0
    statements: str = _serialize_triples_to_nt(data)
    delete_string: str = f"DELETE DATA {{ GRAPH <{graph_iri}> {{ {statements} }} }}"
    return delete_string, num_of_statements


def get_insert_query(graph_iri: URIRef, data: Set) -> Tuple[str, int]:
    num_of_statements: int = len(data)
    if num_of_statements <= 0:
        return "", 0
    statements: str = _serialize_triples_to_nt(data)
    insert_string: str = f"INSERT DATA {{ GRAPH <{graph_iri}> {{ {statements} }} }}"
    return insert_string, num_of_statements


def _compute_graph_changes(entity: AbstractEntity, entity_type: str) -> Tuple[Set, Set, int, int]:
    """
    Computes the triples to insert and delete for an entity.

    Args:
        entity: The entity to analyze
        entity_type: Type of entity ("graph", "prov", or "metadata")

    Returns:
        Tuple of (triples_to_insert, triples_to_delete, added_count, removed_count)
    """
    if entity_type == "prov":
        triples = set(entity.g)
        return triples, set(), len(triples), 0

    to_be_deleted: bool = entity.to_be_deleted
    preexisting_graph = entity.preexisting_graph

    if to_be_deleted:
        preexisting_triples = set(preexisting_graph)
        return set(), preexisting_triples, 0, len(preexisting_triples)

    preexisting_triples = set(preexisting_graph)
    current_triples = set(entity.g)

    if preexisting_triples == current_triples:
        return set(), set(), 0, 0

    removed_triples = preexisting_triples - current_triples
    added_triples = current_triples - preexisting_triples

    return added_triples, removed_triples, len(added_triples), len(removed_triples)


def serialize_graph_to_nquads(triples: Set, graph_iri: URIRef) -> list:
    """
    Serializes RDF triples to N-Quads format.

    Args:
        triples: Set of RDF triples
        graph_iri: Named graph IRI

    Returns:
        List of N-Quad strings (each ending with newline)
    """
    return [f"{s.n3()} {p.n3()} {o.n3()} <{graph_iri}> .\n" for s, p, o in triples]


def get_separated_queries(entity: AbstractEntity, entity_type: str = "graph") -> Tuple[str, str, int, int, Set]:
    """
    Returns separate INSERT and DELETE queries for an entity, plus the insert triples.

    Args:
        entity: The entity to generate queries for
        entity_type: Type of entity ("graph", "prov", or "metadata")

    Returns:
        Tuple of (insert_query, delete_query, added_count, removed_count, insert_triples)
        The insert_triples can be used for direct N-Quads serialization without parsing SPARQL.
    """
    to_insert, to_delete, n_added, n_removed = _compute_graph_changes(entity, entity_type)

    if n_added == 0 and n_removed == 0:
        return "", "", 0, 0, set()

    delete_string = ""
    insert_string = ""

    if n_removed > 0:
        delete_string, _ = get_delete_query(entity.g.identifier, to_delete)

    if n_added > 0:
        insert_string, _ = get_insert_query(entity.g.identifier, to_insert)

    return insert_string, delete_string, n_added, n_removed, to_insert


def get_update_query(entity: AbstractEntity, entity_type: str = "graph") -> Tuple[str, int, int]:
    to_insert, to_delete, n_added, n_removed = _compute_graph_changes(entity, entity_type)

    if n_added == 0 and n_removed == 0:
        return "", 0, 0

    delete_string, _ = get_delete_query(entity.g.identifier, to_delete)
    insert_string, _ = get_insert_query(entity.g.identifier, to_insert)

    if delete_string != "" and insert_string != "":
        return delete_string + '; ' + insert_string, n_added, n_removed
    elif delete_string != "":
        return delete_string, 0, n_removed
    elif insert_string != "":
        return insert_string, n_added, 0
    else:
        return "", 0, 0


#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2016, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.
from __future__ import annotations

from typing import TYPE_CHECKING

from oc_ocdm.decorators import accepts_only

if TYPE_CHECKING:
    from typing import Optional, List
    from rdflib import URIRef
    from oc_ocdm.graph.entities.bibliographic.reference_pointer import ReferencePointer
from oc_ocdm.graph.graph_entity import GraphEntity
from oc_ocdm.graph.entities.bibliographic_entity import BibliographicEntity


class PointerList(BibliographicEntity):
    """Pointer list (short: pl): a textual device (e.g. '[1, 2, 3]' or '[4-9]') which includes a
       number of reference pointers denoting the specific bibliographic references to which
       the list pertains."""

    @accepts_only('pl')
    def merge(self, other: PointerList) -> None:
        """
        The merge operation allows combining two ``PointerList`` entities into a single one,
        by marking the second entity as to be deleted while also copying its data into the current
        ``PointerList``. Moreover, every triple from the containing ``GraphSet`` referring to the second
        entity gets "redirected" to the current entity: **every other reference contained inside a
        different source (e.g. a triplestore) must be manually handled by the user!**

        In case of functional properties, values from the current entity get overwritten
        by those coming from the second entity while, in all other cases, values from the
        second entity are simply appended to those of the current entity. In this context,
        ``rdfs:label`` is considered as a functional property, while ``rdf:type`` is not.

        :param other: The entity which will be marked as to be deleted and whose properties will
         be merged into the current entity.
        :type other: PointerList
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        super(PointerList, self).merge(other)

        content: Optional[str] = self.get_content()
        if content is not None:
            self.has_content(content)

        rp_list: List[ReferencePointer] = other.get_contained_elements()
        for cur_rp in rp_list:
            self.contains_element(cur_rp)

    # HAS POINTER LIST TEXT
    def get_content(self) -> Optional[str]:
        """
        Getter method corresponding to the ``c4o:hasContent`` RDF predicate.

        :return: The requested value if found, None otherwise
        """
        return self._get_literal(GraphEntity.iri_has_content)

    @accepts_only('literal')
    def has_content(self, string: str) -> None:
        """
        Setter method corresponding to the ``c4o:hasContent`` RDF predicate.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        :param string: The value that will be set as the object of the property related to this method
        :type string: str
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self.remove_content()
        self._create_literal(GraphEntity.iri_has_content, string)

    def remove_content(self) -> None:
        """
        Remover method corresponding to the ``c4o:hasContent`` RDF predicate.

        :return: None
        """
        self.g.remove((self.res, GraphEntity.iri_has_content, None))

    # HAS ELEMENT (ReferencePointer)
    def get_contained_elements(self) -> List[ReferencePointer]:
        """
        Getter method corresponding to the ``co:element`` RDF predicate.

        :return: A list containing the requested values if found, None otherwise
        """
        uri_list: List[URIRef] = self._get_multiple_uri_references(GraphEntity.iri_has_element, 'rp')
        result: List[ReferencePointer] = []
        for uri in uri_list:
            result.append(self.g_set.add_rp(self.resp_agent, self.source, uri))
        return result

    @accepts_only('rp')
    def contains_element(self, rp_res: ReferencePointer) -> None:
        """
        Setter method corresponding to the ``co:element`` RDF predicate.

        `The in-text reference pointer that is part of the in-text reference pointer list present at
        a particular location within the body of the citing work.`

        :param rp_res: The value that will be set as the object of the property related to this method
        :type rp_res: ReferencePointer
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self.g.add((self.res, GraphEntity.iri_has_element, rp_res.res))

    @accepts_only('rp')
    def remove_contained_element(self, rp_res: ReferencePointer = None) -> None:
        """
        Remover method corresponding to the ``co:element`` RDF predicate.

        **WARNING: this is a non-functional property, hence, if the parameter
        is None, any existing value will be removed!**

        :param rp_res: If not None, the specific object value that will be removed from the property
         related to this method (defaults to None)
        :type rp_res: ReferencePointer
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        if rp_res is not None:
            self.g.remove((self.res, GraphEntity.iri_has_element, rp_res.res))
        else:
            self.g.remove((self.res, GraphEntity.iri_has_element, None))



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2016, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.
from __future__ import annotations

import re
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Optional
    from rdflib import URIRef

from oc_ocdm.decorators import accepts_only
from oc_ocdm.graph.graph_entity import GraphEntity
from oc_ocdm.support.support import encode_url, is_string_empty


class Identifier(GraphEntity):
    """Identifier (short: id): an external identifier (e.g. DOI, ORCID, PubMedID, Open
       Citation Identifier) associated with the bibliographic entity. Members of this class of
       metadata are themselves given unique corpus identifiers e.g. 'id/0420129'."""

    @accepts_only('id')
    def merge(self, other: Identifier):
        """
        The merge operation allows combining two ``Identifier`` entities into a single one,
        by marking the second entity as to be deleted while also copying its data into the current
        ``Identifier``. Moreover, every triple from the containing ``GraphSet`` referring to the second
        entity gets "redirected" to the current entity: **every other reference contained inside a
        different source (e.g. a triplestore) must be manually handled by the user!**

        In case of functional properties, values from the current entity get overwritten
        by those coming from the second entity while, in all other cases, values from the
        second entity are simply appended to those of the current entity. In this context,
        ``rdfs:label`` is considered as a functional property, while ``rdf:type`` is not.

        :param other: The entity which will be marked as to be deleted and whose properties will
         be merged into the current entity.
        :type other: Identifier
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        super(Identifier, self).merge(other)

        literal_value: Optional[str] = other.get_literal_value()
        scheme: Optional[URIRef] = other.get_scheme()
        if literal_value is not None and scheme is not None:
            self._associate_identifier_with_scheme(literal_value, scheme)

    # HAS LITERAL VALUE and HAS SCHEME
    def get_literal_value(self) -> Optional[str]:
        """
        Getter method corresponding to the ``literal:hasLiteralValue`` RDF predicate.

        :return: The requested value if found, None otherwise
        """
        return self._get_literal(GraphEntity.iri_has_literal_value)

    def get_scheme(self) -> Optional[URIRef]:
        """
        Getter method corresponding to the ``datacite:usesIdentifierScheme`` RDF predicate.

        :return: The requested value if found, None otherwise
        """
        uri: Optional[URIRef] = self._get_uri_reference(GraphEntity.iri_uses_identifier_scheme)
        return uri

    @accepts_only('literal')
    def create_oci(self, string: str) -> None:
        """
        Setter method corresponding to both the ``literal:hasLiteralValue`` and the
        ``datacite:usesIdentifierScheme`` RDF predicate.
        It implicitly sets the object value ``datacite:oci`` for the
        ``datacite:usesIdentifierScheme`` RDF predicate.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        :param string: The value that will be set as the object of the property related to this method
        :type string: str
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self._associate_identifier_with_scheme(string, GraphEntity.iri_oci)

    @accepts_only('literal')
    def create_orcid(self, string: str) -> None:
        """
        Setter method corresponding to both the ``literal:hasLiteralValue`` and the
        ``datacite:usesIdentifierScheme`` RDF predicate.
        It implicitly sets the object value ``datacite:orcid`` for the
        ``datacite:usesIdentifierScheme`` RDF predicate.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        :param string: The value that will be set as the object of the property related to this method
        :type string: str
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self._associate_identifier_with_scheme(string, GraphEntity.iri_orcid)

    @accepts_only('literal')
    def create_openalex(self, string: str) -> None:
        """
        Setter method corresponding to both the ``literal:hasLiteralValue`` and the
        ``datacite:usesIdentifierScheme`` RDF predicate.
        It implicitly sets the object value ``datacite:openalex`` for the
        ``datacite:usesIdentifierScheme`` RDF predicate.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        :param string: The value that will be set as the object of the property related to this method
        :type string: str
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self._associate_identifier_with_scheme(string, GraphEntity.iri_openalex)

    @accepts_only('literal')
    def create_doi(self, string: str) -> None:
        """
        Setter method corresponding to both the ``literal:hasLiteralValue`` and the
        ``datacite:usesIdentifierScheme`` RDF predicate.
        It implicitly sets the object value ``datacite:doi`` for the
        ``datacite:usesIdentifierScheme`` RDF predicate.

        The string gets internally preprocessed by converting it to lowercase
        (e.g. 'DOI:10.1111/HEX.12487' becomes 'doi:10.1111/hex.12487').

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        :param string: The value that will be set as the object of the property related to this method
        :type string: str
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self._associate_identifier_with_scheme(string.lower(), GraphEntity.iri_doi)

    @accepts_only('literal')
    def create_jid(self, string: str) -> None:
        """
        Setter method corresponding to both the ``literal:hasLiteralValue`` and the
        ``datacite:usesIdentifierScheme`` RDF predicate.
        It implicitly sets the object value ``datacite:jid`` for the
        ``datacite:usesIdentifierScheme`` RDF predicate.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        :param string: The value that will be set as the object of the property related to this method
        :type string: str
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self._associate_identifier_with_scheme(string, GraphEntity.iri_jid)

    @accepts_only('literal')
    def create_pmid(self, string: str) -> None:
        """
        Setter method corresponding to both the ``literal:hasLiteralValue`` and the
        ``datacite:usesIdentifierScheme`` RDF predicate.
        It implicitly sets the object value ``datacite:pmid`` for the
        ``datacite:usesIdentifierScheme`` RDF predicate.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        :param string: The value that will be set as the object of the property related to this method
        :type string: str
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self._associate_identifier_with_scheme(string, GraphEntity.iri_pmid)

    @accepts_only('literal')
    def create_pmcid(self, string: str) -> None:
        """
        Setter method corresponding to both the ``literal:hasLiteralValue`` and the
        ``datacite:usesIdentifierScheme`` RDF predicate.
        It implicitly sets the object value ``datacite:pmcid`` for the
        ``datacite:usesIdentifierScheme`` RDF predicate.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        :param string: The value that will be set as the object of the property related to this method
        :type string: str
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self._associate_identifier_with_scheme(string, GraphEntity.iri_pmcid)

    @accepts_only('literal')
    def create_issn(self, string: str) -> None:
        """
        Setter method corresponding to both the ``literal:hasLiteralValue`` and the
        ``datacite:usesIdentifierScheme`` RDF predicate.
        It implicitly sets the object value ``datacite:issn`` for the
        ``datacite:usesIdentifierScheme`` RDF predicate.

        The string gets internally preprocessed by eventually replacing long dashes with short ones
        (e.g. '1522–4501' becomes '1522-4501').

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        :param string: The value that will be set as the object of the property related to this method. **It
          must be a string different from '0000-0000'.**
        :type string: str
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        cur_string = re.sub("–", "-", string)
        if cur_string != "0000-0000":
            self._associate_identifier_with_scheme(string, GraphEntity.iri_issn)

    @accepts_only('literal')
    def create_isbn(self, string: str) -> None:
        """
        Setter method corresponding to both the ``literal:hasLiteralValue`` and the
        ``datacite:usesIdentifierScheme`` RDF predicate.
        It implicitly sets the object value ``datacite:isbn`` for the
        ``datacite:usesIdentifierScheme`` RDF predicate.

        The string gets internally preprocessed by eventually replacing long dashes with short ones
        (e.g. '817525766–0' becomes '817525766-0').

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        :param string: The value that will be set as the object of the property related to this method
        :type string: str
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self._associate_identifier_with_scheme(re.sub("–", "-", string), GraphEntity.iri_isbn)

    @accepts_only('literal')
    def create_url(self, string: str) -> None:
        """
        Setter method corresponding to both the ``literal:hasLiteralValue`` and the
        ``datacite:usesIdentifierScheme`` RDF predicate.
        It implicitly sets the object value ``datacite:url`` for the
        ``datacite:usesIdentifierScheme`` RDF predicate.

        The string gets internally preprocessed both by converting it to lowercase
        (e.g. 'https://OPENCITATIONS.NET/' becomes 'https://opencitations.net/') and by
        applying `URL encoding` on it (e.g. 'https://opencitations.net/file name.txt'
        becomes 'https://opencitations.net/file%20name.txt').

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        :param string: The value that will be set as the object of the property related to this method
        :type string: str
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self._associate_identifier_with_scheme(encode_url(string.lower()), GraphEntity.iri_url)

    @accepts_only('literal')
    def create_xpath(self, string: str) -> None:
        """
        Setter method corresponding to both the ``literal:hasLiteralValue`` and the
        ``datacite:usesIdentifierScheme`` RDF predicate.
        It implicitly sets the object value `datacite:local-resource-identifier-scheme` for the
        ``datacite:usesIdentifierScheme`` RDF predicate.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        :param string: The value that will be set as the object of the property related to this method
        :type string: str
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self._associate_identifier_with_scheme(string, GraphEntity.iri_xpath)

    @accepts_only('literal')
    def create_intrepid(self, string: str) -> None:
        """
        Setter method corresponding to both the ``literal:hasLiteralValue`` and the
        ``datacite:usesIdentifierScheme`` RDF predicate.
        It implicitly sets the object value ``datacite:intrepid`` for the
        ``datacite:usesIdentifierScheme`` RDF predicate.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        :param string: The value that will be set as the object of the property related to this method
        :type string: str
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self._associate_identifier_with_scheme(string, GraphEntity.iri_intrepid)

    @accepts_only('literal')
    def create_xmlid(self, string: str) -> None:
        """
        Setter method corresponding to both the ``literal:hasLiteralValue`` and the
        ``datacite:usesIdentifierScheme`` RDF predicate.
        It implicitly sets the object value `datacite:local-resource-identifier-scheme` for the
        ``datacite:usesIdentifierScheme`` RDF predicate.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        :param string: The value that will be set as the object of the property related to this method
        :type string: str
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self._associate_identifier_with_scheme(string, GraphEntity.iri_xmlid)

    @accepts_only('literal')
    def create_wikidata(self, string: str) -> None:
        """
        Setter method corresponding to both the ``literal:hasLiteralValue`` and the
        ``datacite:usesIdentifierScheme`` RDF predicate.
        It implicitly sets the object value ``datacite:wikidata`` for the
        ``datacite:usesIdentifierScheme`` RDF predicate.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        :param string: The value that will be set as the object of the property related to this method
        :type string: str
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self._associate_identifier_with_scheme(string, GraphEntity.iri_wikidata)

    @accepts_only('literal')
    def create_wikipedia(self, string: str) -> None:
        """
        Setter method corresponding to both the ``literal:hasLiteralValue`` and the
        ``datacite:usesIdentifierScheme`` RDF predicate.
        It implicitly sets the object value ``datacite:wikipedia`` for the
        ``datacite:usesIdentifierScheme`` RDF predicate.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        :param string: The value that will be set as the object of the property related to this method
        :type string: str
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self._associate_identifier_with_scheme(string, GraphEntity.iri_wikipedia)

    @accepts_only('literal')
    def create_arxiv(self, string: str) -> None:
        """
        Setter method corresponding to both the ``literal:hasLiteralValue`` and the
        ``datacite:usesIdentifierScheme`` RDF predicate.
        It implicitly sets the object value ``datacite:crossref`` for the
        ``datacite:usesIdentifierScheme`` RDF predicate.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        :param string: The value that will be set as the object of the property related to this method
        :type string: str
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self._associate_identifier_with_scheme(string, GraphEntity.iri_arxiv)

    @accepts_only('literal')
    def create_crossref(self, string: str) -> None:
        """
        Setter method corresponding to both the ``literal:hasLiteralValue`` and the
        ``datacite:usesIdentifierScheme`` RDF predicate.
        It implicitly sets the object value ``datacite:crossref`` for the
        ``datacite:usesIdentifierScheme`` RDF predicate.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        :param string: The value that will be set as the object of the property related to this method
        :type string: str
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self._associate_identifier_with_scheme(string, GraphEntity.iri_crossref)

    @accepts_only('literal')
    def create_datacite(self, string: str) -> None:
        """
        Setter method corresponding to both the ``literal:hasLiteralValue`` and the
        ``datacite:usesIdentifierScheme`` RDF predicate.
        It implicitly sets the object value ``datacite:datacite`` for the
        ``datacite:usesIdentifierScheme`` RDF predicate.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        :param string: The value that will be set as the object of the property related to this method
        :type string: str
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self._associate_identifier_with_scheme(string, GraphEntity.iri_datacite)

    @accepts_only('literal')
    def create_viaf(self, string: str) -> None:
        """
        Setter method corresponding to both the ``literal:hasLiteralValue`` and the
        ``datacite:usesIdentifierScheme`` RDF predicate.
        It implicitly sets the object value ``datacite:viaf`` for the
        ``datacite:usesIdentifierScheme`` RDF predicate.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        :param string: The value that will be set as the object of the property related to this method
        :type string: str
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self._associate_identifier_with_scheme(string, GraphEntity.iri_viaf)

    def _associate_identifier_with_scheme(self, string: str, id_type: URIRef) -> None:
        if not is_string_empty(string):
            self.remove_identifier_with_scheme()
            self._create_literal(GraphEntity.iri_has_literal_value, string)
            self.g.add((self.res, GraphEntity.iri_uses_identifier_scheme, id_type))

    def remove_identifier_with_scheme(self) -> None:
        """
        Remover method corresponding to both the ``literal:hasLiteralValue`` and the
        ``datacite:usesIdentifierScheme`` RDF predicate.

        :return: None
        """
        self.g.remove((self.res, GraphEntity.iri_has_literal_value, None))
        self.g.remove((self.res, GraphEntity.iri_uses_identifier_scheme, None))
